/*
 * Copyright (C) 2010 Timo Kokkonen <kaapeli@itanic.dy.fi>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <libgen.h>
#include <errno.h>
#include <string.h>

#include "pidlib.h"

char *get_name_by_pid(int pid)
{
	static int last_pid;
	static char cmdline[128];
	static char *bname;

	if (last_pid == pid)
		return bname;

	if (read_cmdline(pid, pid, cmdline, sizeof(cmdline))) {
		bname = NULL;
		return NULL;
	}

	bname = basename(cmdline);

	last_pid = pid;
	return bname;
}

static int parse_pid(DIR **dir)
{
	struct dirent *dirent;
	int error;

restart:
	dirent = readdir(*dir);
	if (!dirent) {
		if (errno == 0) {
			closedir(*dir);
			*dir = NULL;
			return 0;
		}
		error = errno;
		printf("Failed to read /proc directory: %s\n", strerror(error));
		return -1;
	}

	if (dirent->d_name[0] < '0' || dirent->d_name[0] > '9')
		goto restart;

	return atoi(dirent->d_name);
}

static int opendir_check(DIR **dir, const char *path)
{
	int error;

	if (!*dir) {
		*dir = opendir(path);
		if (!dir) {
			error = errno;
			fprintf(stderr, "Failed to open %s directory: %s\n",
				path, strerror(error));
			return -1;
		}
	}

	return 0;
}

int get_next_tid(int pid, DIR **dir)
{
	if (*dir == NULL) {
		char path[64];

		snprintf(path, sizeof(path), "/proc/%d/task/", pid);
		if (opendir_check(dir, path))
			return -1;
	}

	return parse_pid(dir);
}

int get_next_pid(DIR **dir)
{
	if (opendir_check(dir, "/proc"))
		return -1;

	return parse_pid(dir);
}

int get_next_pid_by_name(DIR **dir, char *name)
{
	int pid;
	char *pname;

	if (opendir_check(dir, "/proc"))
		return -1;

	while (1) {
		pid = parse_pid(dir);
		if (pid <= 0)
			break;

		pname = get_name_by_pid(pid);
		if (pname == NULL)
			continue;
		if (strcmp(pname, name))
			continue;

		return pid;
	}

	return 0;
}

int read_cmdline(int pid, int tid, char *cmdline, size_t len)
{
	FILE *file;
	char path[512];
	int ret;

	snprintf(path, sizeof(path), "/proc/%d/task/%d/cmdline", pid, tid);
	file = fopen(path, "rb");

	if (!file)
		return -1;

	ret = fread(cmdline, 1, len, file);
	if (ret > 0)
		cmdline[ret - 1] = 0;
	fclose(file);

	return ret > 0 ? 0 : -1;
}

static int pid_is_ok(int pid)
{
	struct stat statt;
	char path[64];

	snprintf(path, sizeof(path), "/proc/%d/", pid);

	if (stat(path, &statt))
		return 0;

	return pid;
}

/*
 * Test whether a pid string number exists in the system as a pid
 * number
 */
int pidstr_is_ok(const char *str)
{
	int i, pid;

	/* Test that it is all digits */
	for (i = 0; i < strlen(str); i++)
		if (str[i] < '0' || str[i] > '9')
			return 0;

	pid = atoi(str);

	return pid_is_ok(pid);
}
